import os

from pydantic import BaseModel
from sanic import Blueprint
from sanic.views import HTTPMethodView

from os_sanic.utils import load_class, normalize_slash


class URIModel(BaseModel):
    uri: str

    class Config:
        allow_extra = True


class RouteCfg(URIModel):
    view_class: str


class StaticCfg(URIModel):
    file_or_directory: str


def adapt_uri(url_prefix, uri):
    real_uri = normalize_slash(uri, with_prefix_slash=False)

    if url_prefix.endswith('/') and real_uri.startswith('/'):
        real_uri = real_uri[1:]

    return real_uri


def new_blueprint(application):
    _prefix = application.app_cfg.prefix

    if not _prefix:
        _prefix = f'/{application.name}'

    prefix = normalize_slash(_prefix)

    if not prefix == _prefix:
        application.logger.warn(
            f'Normalize prefix from \'{_prefix}\' to \'{prefix}\'')

    return Blueprint(application.name, url_prefix=prefix)


def load_route(application, blueprint, route_cfg, user_cfg):
    config = route_cfg.copy(update=user_cfg.copy(
        exclude=set(['view_class'])).dict())

    package = None
    if config.view_class.startswith('.'):
        package = application.app_cfg.package
    view_cls = load_class(
        config.view_class, HTTPMethodView, package=package)

    view_cls.application = application
    view_cls.config = config

    real_uri = adapt_uri(blueprint.url_prefix, config.uri)
    view = blueprint.add_route(view_cls.as_view(), real_uri)

    pth = blueprint.url_prefix+real_uri
    application.logger.debug(f'Load route, {pth} {view.view_class}')


def load_routes(application, blueprint):
    user_cfgs = dict([(cfg['uri'], URIModel(**cfg))
                      for cfg in application.user_cfg.ROUTES if 'uri' in cfg])

    for cfg in application.core_cfg.ROUTES:
        try:
            if isinstance(cfg, tuple):
                cfg = dict(uri=cfg[0], view_class=cfg[1])
            route_cfg = RouteCfg(**cfg)

            load_route(application, blueprint, route_cfg, user_cfgs.get(
                route_cfg.uri, route_cfg))
        except Exception as e:
            application.logger.error(f'Load route error, {e}, {cfg}')


def load_static(application, blueprint, static_cfg, user_cfg):
    config = static_cfg.copy(update=user_cfg.dict())
    real_uri = adapt_uri(blueprint.url_prefix, config.uri)

    blueprint.static(real_uri, os.path.abspath(os.path.join(
        application.runtime_path, config.file_or_directory)),
        **config.dict(exclude=set(['uri', 'file_or_directory'])))

    static = blueprint.statics[-1]
    pth = blueprint.url_prefix + static.uri
    application.logger.debug(
        f'Load static, {pth} {blueprint.statics[-1]}')


def load_statics(application, blueprint):
    user_cfgs = dict([(cfg['uri'], URIModel(**cfg))
                      for cfg in application.user_cfg.STATICS if 'uri' in cfg])

    for cfg in application.core_cfg.STATICS:
        try:
            static_cfg = StaticCfg(**cfg)
            load_static(application, blueprint, static_cfg, user_cfgs.get(
                static_cfg.uri, static_cfg))
        except Exception as e:
            application.logger.error(f'Load static error, {e}, {cfg}')


def create(application):
    blueprint = new_blueprint(application)
    load_routes(application, blueprint)
    load_statics(application, blueprint)
    application.sanic.blueprint(blueprint)
    return blueprint



import os
from typing import List, Union

from pydantic import BaseSettings, PositiveInt
from sanic.config import Config

from os_sanic.utils import LogLevel

SANIC_ENV_PREFIX = 'OS_SANIC_'


class DefaultConfig(BaseSettings):

    LOGO: str = None

    DEBUG = False

    LOG_LEVEL: LogLevel = 'INFO'

    ACCESS_LOG = True

    KEEP_ALIVE = True

    HOST = '127.0.0.1'

    PORT: PositiveInt = 8001

    INSTALLED_APPS: List[Union[str, dict]] = []

    class Config:
        env_prefix = SANIC_ENV_PREFIX
        ignore_extra = True


def create(params=None, load_env=True):
    config = Config(load_env=False)
    config.update(dict([(k, v.default)
                        for k, v in DefaultConfig.__fields__.items()]))

    if load_env:
        if isinstance(load_env, bool):
            load_env = SANIC_ENV_PREFIX
        config.load_environment_vars(prefix=load_env)

    if params:
        config.update(params)
    DefaultConfig.validate(config)
    return config



EXTENSIONS = [
    {
        'name': 'case01',
        'extension_class': '.extension.Case01Extension'
    }
]

ROUTES = [
    ('/', '.view.Case01View'),
    {
        'uri': '/config',
        'view_class': '.view.Case02View',
        'key01': 'value01',
    }
]



import json
import os
from collections import OrderedDict

import click

from os_sanic.commands.project.run import create_server


def server_info(server):
    click.echo('Server:')
    out = json.dumps(server.sanic.config, indent=4)
    click.echo(f'{out}\n')


def apps_info(application_manager):
    apps_info = []
    for app in application_manager.apps:
        info = app_info(app)
        apps_info.append(info)

    if apps_info:
        out = json.dumps(apps_info, indent=4)
        click.echo('Applications:')
        click.echo(out)


def extensions_info(app):
    exts_info = []

    for extension in app.extension_manager.extensions:
        exts_info.append(extension.config.copy(
            update={'extension_class': str(extension.__class__)}).dict())
    return exts_info


def routes_info(app):
    routes_info = []
    for route in app.blueprint.routes:
        view_cls = route.handler.view_class
        routes_info.append(view_cls.config.copy(
            update={'uri': route.uri, 'view_class': str(view_cls)}).dict())
    return routes_info


def statics_info(app):
    statics_info = []
    for static in app.blueprint.statics:
        static_info = OrderedDict()
        static_info['uri'] = static.uri
        static_info['file_or_directory'] = static.file_or_directory
        static_info.update(static.kwargs)
        statics_info.append(static_info)

    return statics_info


def app_info(app):
    info = app.app_cfg.copy(
        update={'prefix': app.blueprint.url_prefix}).dict()

    for method in (extensions_info, routes_info, statics_info):
        i=method(app)
        if i:
            info[method.__name__[:-5]]=i

    return info


@click.command()
@click.option('-c', '--config-file',
              default = 'config.py',
              show_default = True,
              type = click.File(mode='r'),
              help = 'Config file')
@click.pass_context
def cli(ctx, config_file):
    '''Show config details.'''

    ctx.ensure_object(dict)

    server=create_server(ctx.obj['app'], config_file = config_file)
    server_info(server)
    apps_info(server.application_manager)



from collections import OrderedDict
from functools import partial
from inspect import isawaitable

from os_sanic.utils import NamedModel, load_class
from os_sanic.workflow import Workflowable


class ExtensionCfg(NamedModel):
    extension_class: str


class Extension(Workflowable):

    def __init__(self, application, config):
        self.application = application
        self.config = config
        self.logger = application.get_logger(config.name)

    @staticmethod
    def create(application, ext_cfg, user_cfg):
        config = ext_cfg.copy(update=user_cfg.copy(
            exclude=set(['extension_class'])).dict())

        package = None
        if config.extension_class.startswith('.'):
            package = application.app_cfg.package
        cls = load_class(config.extension_class, Extension, package=package)

        return cls(application, config)


class ExtensionManager(Workflowable):
    def __init__(self, application):
        self.application = application
        self.logger = application.get_logger(self.__class__.__name__)
        self._extensions = OrderedDict()
        self._load_extensions()
        [setattr(self, method, partial(self.__call, method))
         for method in ('setup', 'cleanup')]

    def _load_extensions(self):
        user_cfgs = dict([(cfg['name'], NamedModel(**cfg))
                          for cfg in self.application.user_cfg.EXTENSIONS if 'name' in cfg])

        for cfg in self.application.core_cfg.EXTENSIONS:
            try:
                ext_cfg = ExtensionCfg(**cfg)
                self._load_extension(
                    ext_cfg, user_cfgs.get(ext_cfg.name, ext_cfg))
            except Exception as e:
                self.logger.error(f'Load extension fail {e}, {ext_cfg}')

    def get_extension(self, name):
        return self._extensions[name]

    @property
    def extensions(self):
        return self._extensions.values()

    async def __call(self, method):
        iter = self._extensions.keys()
        if method == 'cleanup':
            iter = sorted(iter, reverse=True)

        for key in iter:
            ext = self._extensions[key]
            try:
                r = getattr(ext, method)()
                if isawaitable(r):
                    await r

            except Exception as e:
                self.logger.error(f'Extension error {key}.{method}, {e}')

    def _load_extension(self, ext_cfg, user_cfg):
        name = ext_cfg.name
        if name in self._extensions:
            self.logger.warn(f'Extension already exists, {name}')
            return
        extension = Extension.create(
            self.application, ext_cfg, user_cfg)
        self._extensions[name] = extension
        self.logger.debug(f'Load extension, {name} {extension.__class__}')



import inspect
import os
from collections import OrderedDict
from functools import partial
from importlib import import_module
from inspect import isawaitable
from typing import List

from pydantic import BaseModel

from os_sanic.extension import ExtensionCfg, ExtensionManager
from os_sanic.log import getLogger
from os_sanic.utils import NamedModel, load_module_from_pyfile
from os_sanic.workflow import Workflowable
from os_sanic import blueprint


class AppCfg(NamedModel):
    package: str
    prefix: str = None
    config: str = None


class ApplicationCfg(BaseModel):
    EXTENSIONS: List = []
    ROUTES: List = []
    STATICS: List = []

    class Config:
        ignore_extra = True


class Context(BaseModel):
    app_cfg: AppCfg
    core_cfg: ApplicationCfg
    user_cfg: ApplicationCfg
    runtime_path: str


class Application(Workflowable):

    def __init__(self, application_manager, context):
        self.context = context
        self.application_manager = application_manager

        [setattr(self, method, partial(self.__call, method))
         for method in ('setup', 'cleanup')]

        [setattr(self, field, getattr(context, field))
         for field in context.fields.keys()]

        self.logger = getLogger(f'App.{self.name}')

        self.extension_manager = ExtensionManager(self)
        self.blueprint = blueprint.create(self)

    @property
    def name(self):
        return self.app_cfg.name

    @property
    def sanic(self):
        return self.application_manager.sanic

    def get_extension(self, extension_path):
        if '.' not in extension_path:
            extension_path = '.'.join((self.name, extension_path))

        return self.application_manager.get_extension(extension_path)

    def get_logger(self, tag):
        return getLogger(f'App.{self.name}.{tag}')

    async def __call(self, method):
        try:
            await getattr(self.extension_manager, method)()
        except Exception as e:
            self.logger.error(f'Method error {method}, {e}')

    @staticmethod
    def create(application_manager, app_cfg):

        app_module = import_module('.'.join((app_cfg.package, 'app')))
        runtime_path = os.path.dirname(app_module.__file__)
        core_cfg = ApplicationCfg.parse_obj(app_module.__dict__)

        user_cfg = ApplicationCfg()
        if app_cfg.config:
            f = os.path.join(os.getcwd(), app_cfg.config)
            user_cfg = ApplicationCfg.parse_obj(
                load_module_from_pyfile(f).__dict__)
            runtime_path = os.path.dirname(f)

        context = Context(
            app_cfg=app_cfg,
            core_cfg=core_cfg,
            user_cfg=user_cfg,
            runtime_path=runtime_path,
        )

        return Application(application_manager, context)


class ApplicationManager(Workflowable):

    def __init__(self, sanic):
        self.sanic = sanic
        self.logger = getLogger(self.__class__.__name__)
        self._apps = OrderedDict()
        self._load_apps()
        [setattr(self, method, partial(self.__call, method))
         for method in ('setup', 'cleanup')]

    async def __call(self, method):
        iter = self._apps.keys()
        if method == 'cleanup':
            iter = sorted(iter, reverse=True)

        for key in iter:
            app = self._apps[key]
            try:
                await getattr(app, method)()
            except Exception as e:
                self.logger.error(f'Application error {key}.{method}, {e}')

    def get_app(self, name):
        return self._apps[name]

    def get_extension(self, extension_path):
        app_name, extension_name = extension_path.split('.')
        app = self.get_app(app_name)
        return app.extension_manager.get_extension(extension_name)

    @property
    def apps(self):
        return self._apps.values()

    def _load_apps(self):
        for cfg in self.sanic.config.get('INSTALLED_APPS', []):
            try:
                self._load_app(cfg)
            except Exception as e:
                self.logger.error(f'Load app fail, {e}, {cfg}')

    def _load_app(self, cfg):

        if isinstance(cfg, str):
            cfg = dict(package=cfg)

        cfg.setdefault('name', cfg['package'].split('.')[-1])
        name = cfg['name']
        if name in self._apps:
            self.logger.warn(f'Skip, app already exists, {name}')
            return

        app_cfg = AppCfg(**cfg)
        self.logger.debug(
            f'Load app, {name} <package \'{app_cfg.package}>\'')

        app = Application.create(self, app_cfg)
        self._apps[name] = app
